-----------------------------------------------------------------------------------------------------------------------------------
-- Engineer			: 		Dennis van Elteren, Maarssen, The Netherlands
-- 
-- Create Date		:    	February 24th of 2015
-- Design Name		: 		My FPGA vision system - the ILI9325 TFT driver
-- Module Name		:    	ili9325 - Behavioral 
-- Target Devices	: 		xc3s200,xc6slx9
-- Tool versions	: 		ISE14.7
-- Description		: 		This vhdl code initialises an IL9325 based 320x240 RGB display RGB565 
--								(the display interface was in 16-bit mode (see IM[3:0] in ILI9325 datasheet)
--								It then waits for bytes received on its input: CAMERA_DATA_IN
--								on each falling edge of NEW_FRAME_FOR_TFT one byte will be sent to the TFT screen (2 bytes form 1 pixel)
--								NOTE: pixelclock is equal to crystalclock_50MHZ divided by 2, this is why the complete state machine runs on the same clock
--								Pixelclock is an input in the VHDL file but is not used at the moment
--
-- Current Rev		: 		1.5
-- History			: 		1.0	- 	File Created
--								1.1	-	After display init this version loops the TFT color from 0x0000 to 0xffff
--								1.2	-	Goes further with the V1.0 code. Colors the TFT yellow after initialisation
--											Added the statemachine which displays the frames from the OV7670 camera.
--											This version has the OV7670 interface implementation which after initialisation displays
--											the camera data (what the camera sees), triggered on the OV7670 VSYNC line. On toplevel the 
--											camera is "simulated": VSYNC is a 250Hz derivative of the 50Mhz clk, HREF is constantly high
--											because in planahead a pullup is placed on its pin. OV7670_DATABUS is connected to the s3 board
--											8 input switches: all switches off simulate WHITE, all on simulate BLACK
--								1.3	-	No changes to this code but the OV7670 databus is now not connected to the s3 switches but to the real camera
--								1.4	-	Driver is connected to ov7670 simulator (VHDL code) and the simulated testpattern is displayed ALMOST correctly!
--								1.5	-	The driver now is working correctly with an ov7670 camera in QVGA RGB565 mode with 25MHz Clock!
--										-	The driver now takes 16-bit words instead of 8-bit, from an extra code block between ILI9325 driver and OV7670
--											camera. Also added a new RESET mechanism
------------------------------------------------------------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL; --for casting std_logic_vector to unsigned

entity ili9325_16bitcameradata is
    Generic (Crystal_frequency : integer   := 25000000; -- in Hz
				 totalarraysize    : integer   := 101  );   -- size of the initialisation array (101 values for ili9325 example)
	 Port ( DATA_TO_ILI9325 			: out  STD_LOGIC_VECTOR (15 downto 0) := X"0000";
			  DATA_COMING_FROM_ILI9325 : in   STD_LOGIC_VECTOR (15 downto 0);
			  CAMERA_DATA_IN 				: in   STD_LOGIC_VECTOR (15 downto 0); --interfaces to an image processing code block
			  CLOCK_IN 						: in   STD_LOGIC;
			  NEW_FRAME_FOR_TFT 			: in   STD_LOGIC; --interfaces to OV7670 VSYNC line, indicates the start and stop of a frame
			  PIXELCLOCK					: in 	 STD_LOGIC; --interfaces to OV7670 pixelclock output, on this clk the pixels are put out
			  HREF_FROM_OV7670			: in	 STD_LOGIC; --interfaces to OV7670 HREF output which indicates the lines in a frame
			  NOT_RESET_IN					: in 	 STD_LOGIC; --to reset the block
           CS_ILI9325 					: out  STD_LOGIC := '1'; --CS,RS,WR,RD and RESET are active-low so startup with inactive
           RS_ILI9325 					: out  STD_LOGIC := '1';
           WR_ILI9325 					: out  STD_LOGIC := '1';
           RD_ILI9325 					: out  STD_LOGIC := '1';
           RESET_ILI9325 				: out  STD_LOGIC := '1';
			  EN_BACKLIGHT 				: out  STD_LOGIC := '0';
			  ILI9325_READY 				: out  STD_LOGIC := '0');
end ili9325_16bitcameradata;

architecture Behavioral of ili9325_16bitcameradata is
	
	--signals for the state machine
	signal delaytime_counter 				: std_logic_vector(31 downto 0) := std_logic_vector(to_unsigned((Crystal_frequency/100), 32)); --this waits 5mS
	signal total_number_of_pixels			: std_logic_vector(31 downto 0) := std_logic_vector(to_unsigned(76800, 32)); --total nr of ili9325 TFT pixels is 76800
	signal total_number_of_pixelbytes	: std_logic_vector(31 downto 0) := std_logic_vector(to_unsigned(153600, 32)); --total nr of pixelbytes coming from the OV7670
	signal color								: std_logic_vector(15 downto 0) := X"0000"; --color initially is white, after the TFT is initialised it will be colored yellow
	signal ili9325_is_initialised 		: std_logic :=  '0';
	
	--types for the state machines
	type ili9325_states is (st_reset,st_prepare_cs,st_prepare_rs_addressmode,st_prepare_write_address,st_write_address,st_prepare_rs_datamode,st_prepare_wr,st_prepare_write_data,st_write_data,
				 					 st_cs_idle_after_write_data,st_prepare_setXcursor_cs,st_prepare_setXcursor_rs_addresmode,st_prepare_setXcursor_wr,st_write_setXcursor_cmd,
									 st_prepare_setXcursor_rs_datamode,st_prepare_write_setXcursordata,st_prepare_setXcursor_data_wr,st_write_setXcursor_data,st_idle_after_setXcursor_data,
									 st_prepare_setYcursor_cs,st_prepare_setYcursor_rs_addresmode,st_prepare_setYcursor_wr,st_write_setYcursor_cmd,st_prepare_setYcursor_rs_datamode,
									 st_prepare_write_setYcursordata,st_prepare_setYcursor_data_wr,st_write_setYcursor_data,st_idle_after_setYcursor_data,st_prepare_select_gram_cs,
									 st_prepare_select_gram_rs,st_prepare_select_gram_wr,st_select_gram,st_cs_idle_after_select_gram,st_prepare_clearscreen_cs,st_prepare_clearscreen_rs,st_prepare_clearscreen_wr,
									 st_clearscreen,st_cs_idle_after_clearscreen,st_delay,st_idle,st_showframe_prepare_setXcursor_cs,st_showframe_prepare_setXcursor_rs_addresmode,
									 st_showframe_prepare_setXcursor_wr,st_showframe_write_setXcursor_cmd,st_showframe_prepare_setXcursor_rs_datamode,st_showframe_prepare_write_setXcursordata,st_showframe_prepare_setXcursor_data_wr,
									 st_showframe_write_setXcursor_data,st_showframe_idle_after_setXcursor_data,st_showframe_prepare_setYcursor_cs,st_showframe_prepare_setYcursor_rs_addresmode,st_showframe_prepare_setYcursor_wr,
									 st_showframe_write_setYcursor_cmd,st_showframe_prepare_setYcursor_rs_datamode,st_showframe_prepare_write_setYcursordata,st_showframe_prepare_setYcursor_data_wr,st_showframe_write_setYcursor_data,
									 st_showframe_idle_after_setYcursor_data,st_showframe_prepare_select_gram_cs,st_showframe_prepare_select_gram_rs,st_showframe_prepare_select_gram_wr,st_showframe_select_gram,
									 st_data2display_pixel_buffering_and_prepare_rs_and_cs,st_data2display_pixel_buffering_and_prepare_wr,st_data2display_write_pixel,st_wait_for_frame_completion); 
									 
	
	-- Register to store the current state(s) or nextstate to go to
	signal ili9325_state : ili9325_states := st_reset; --intial state 
	signal nextstate : ili9325_states := st_reset; --this signal will be used to jump to the next desired state from a delay state so that this delay state can be used anywhere in the state machine
	--signal ili9325_state : ili9325_states := st_idle; --use this when you want to speedup the simulation by starting after TFT init
	
	--array and signals to initialise the ili9325 tft
	type initarray is array (0 to totalarraysize) of std_logic_vector(15 downto 0); 
	signal array_with_ili9325_initvalues : initarray;   --instantiate 1 array of the previously defined type
	signal index : integer range 0 to totalarraysize;
	
	--signals to store temporary data during the running of operating state machine
	signal pixelbuffer1	: std_logic_vector(15 downto 0) := X"0000";
	signal pixelbuffer2	: std_logic_vector(15 downto 0) := X"0000";



	
begin
	--the array "array_with_ili9325_initvalues" containing the registers initvalues is below this state-machine
	
	ili9325_statemachine :	process(CLOCK_IN,NEW_FRAME_FOR_TFT)
	begin
		if rising_edge(CLOCK_IN) then --the whole statemachine responds on rising edge because OV7670 outputs data on rising edge and for ILI9325 it does not matter
			case ili9325_state is
				when st_reset => --wait for a number of clk cycles (power on reset)
					RESET_ILI9325 <= '0'; --reset ili9325
					EN_BACKLIGHT <= '0'; --backlight off
					if (NOT_RESET_IN = '1') then 
						RESET_ILI9325 <= '1'; --end of reset period
						DATA_TO_ILI9325 <= X"0000";
						RD_ILI9325 <= '1';
						ili9325_is_initialised <=  '0';
						ili9325_state <= st_prepare_cs;
					end if;					
					
				when st_prepare_cs => 
					CS_ILI9325 <= '0'; --select the ili9325 chip
					DATA_TO_ILI9325 <= array_with_ili9325_initvalues(index);
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					ili9325_state <= st_delay; 
					nextstate <= st_prepare_rs_addressmode;
					
				when st_prepare_rs_addressmode =>
					RS_ILI9325 <= '0'; --prepare RS pulse for write address
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					ili9325_state <= st_delay; 
					nextstate <= st_prepare_write_address;
					
				when st_prepare_write_address =>
					WR_ILI9325 <= '0'; --prepare the WR pulse
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					ili9325_state <= st_delay; 
					nextstate <= st_write_address;
					
					
				when st_write_address =>
					WR_ILI9325 <= '1'; --write the address
					ili9325_state <= st_delay;
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait					
					nextstate <= st_prepare_rs_datamode;
					
				when st_prepare_rs_datamode =>
					RS_ILI9325 <= '1'; --prepare RS pulse for write data
					index <= index + 1; --move on to the data bytes in the array
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_write_data;
					
				when st_prepare_write_data =>
					DATA_TO_ILI9325 <= array_with_ili9325_initvalues(index);
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					ili9325_state <= st_delay; 
					nextstate <= st_prepare_wr;
					
				when st_prepare_wr =>
					WR_ILI9325 <= '0'; --prepare the WR pulse
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_write_data;
					
				when st_write_data =>
					WR_ILI9325 <= '1'; --write the data
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_cs_idle_after_write_data;
					
				when st_cs_idle_after_write_data =>
					CS_ILI9325 <= '1'; --deselect the ili9325 chip
					if(index = totalarraysize) then --is the complete array sent out? then we are done
						EN_BACKLIGHT <= '1'; --backlight on
						ili9325_state <= st_delay; 
						delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
						nextstate <= st_prepare_setXcursor_cs;
					else
						delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/200), 32)); --not done, prepare a 5ms delaytime (500000 clock cyles)
						ili9325_state <= st_delay;  --5ms is waited before writing the next address and data bytes
						nextstate <= st_prepare_cs;
						index <= index + 1;
					end if;
					
				when st_prepare_setXcursor_cs =>	
					CS_ILI9325 <= '0'; --select the ili9325 chip
					DATA_TO_ILI9325 <= X"0020";
					ili9325_state <= st_delay;
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_setXcursor_rs_addresmode;
					
				when st_prepare_setXcursor_rs_addresmode =>	--set RS line for address/cmd mode
					RS_ILI9325 <= '0'; --prepare RS pulse for write address
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_setXcursor_wr;
					
				when st_prepare_setXcursor_wr =>	--prepare the wr line
					WR_ILI9325 <= '0'; --prepare the WR pulse
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_write_setXcursor_cmd;
					
				when st_write_setXcursor_cmd => --now write the command/address
					WR_ILI9325 <= '1'; --write the address
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_setXcursor_rs_datamode;	
					
				when st_prepare_setXcursor_rs_datamode => --prepare RS line for datamode
					RS_ILI9325 <= '1'; --prepare RS pulse for write data
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_write_setXcursordata;
					
				when st_prepare_write_setXcursordata => --prepare the data to send
					DATA_TO_ILI9325 <= X"0000";
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_setXcursor_data_wr;
					
				when st_prepare_setXcursor_data_wr =>
					WR_ILI9325 <= '0'; --prepare the WR pulse
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_write_setXcursor_data;
					
				when st_write_setXcursor_data =>	
					WR_ILI9325 <= '1'; --write the data
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_idle_after_setXcursor_data;
					
				when st_idle_after_setXcursor_data =>
					CS_ILI9325 <= '1'; --deselect the ili9325 chip
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_setYcursor_cs;
				

				when st_prepare_setYcursor_cs =>	
					CS_ILI9325 <= '0'; --select the ili9325 chip
					DATA_TO_ILI9325 <= X"0021";
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_setYcursor_rs_addresmode;
					
				when st_prepare_setYcursor_rs_addresmode =>	--set RS line for address/cmd mode
					RS_ILI9325 <= '0'; --prepare RS pulse for write address
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_setYcursor_wr;
					
				when st_prepare_setYcursor_wr =>	--prepare the wr line
					WR_ILI9325 <= '0'; --prepare the WR pulse
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_write_setYcursor_cmd;
					
				when st_write_setYcursor_cmd => --now write the command/address
					WR_ILI9325 <= '1'; --write the address
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_setYcursor_rs_datamode;	
					
				when st_prepare_setYcursor_rs_datamode => --prepare RS line for datamode
					RS_ILI9325 <= '1'; --prepare RS pulse for write data
					ili9325_state <= st_delay; delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_write_setYcursordata;
					
				when st_prepare_write_setYcursordata => --prepare the data to send
					DATA_TO_ILI9325 <= X"0000";
					ili9325_state <= st_delay;
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait					
					nextstate <= st_prepare_setYcursor_data_wr;
					
				when st_prepare_setYcursor_data_wr =>
					WR_ILI9325 <= '0'; --prepare the WR pulse
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_write_setYcursor_data;
					
				when st_write_setYcursor_data =>	
					WR_ILI9325 <= '1'; --write the data
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_idle_after_setYcursor_data;
					
				when st_idle_after_setYcursor_data =>
					CS_ILI9325 <= '1'; --deselect the ili9325 chip
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_select_gram_cs; --next state will prepare for a clear-screen of the TFT and put it in a certain color
					
					
				when st_prepare_select_gram_cs =>
					CS_ILI9325 <= '0'; --select the ili9325 chip
					DATA_TO_ILI9325 <= X"0022"; --select GRAM register;
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_select_gram_rs;
					
				when st_prepare_select_gram_rs =>
					RS_ILI9325 <= '0'; --prepare RS pulse for write address
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_select_gram_wr;
					
				when st_prepare_select_gram_wr =>
					WR_ILI9325 <= '0'; --prepare the WR pulse
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_select_gram;


				when st_select_gram =>
					WR_ILI9325 <= '1'; --write the address
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_cs_idle_after_select_gram;
					
				when st_cs_idle_after_select_gram =>
					CS_ILI9325 <= '1'; --deselect the ili9325 chip
					color <= X"FFE0"; --prepare the color to give to the screen after init is finished
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_clearscreen_cs;
					
					
				when st_prepare_clearscreen_cs =>
					CS_ILI9325 <= '0'; --select the ili9325 chip
					DATA_TO_ILI9325 <= color; --select GRAM register;
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_clearscreen_rs;
					
				when st_prepare_clearscreen_rs =>
					RS_ILI9325 <= '1'; --prepare RS pulse for write address
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_prepare_clearscreen_wr;
					
				when st_prepare_clearscreen_wr =>
					WR_ILI9325 <= '0'; --prepare the WR pulse
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_clearscreen;	
					
				when st_clearscreen =>
					WR_ILI9325 <= '1'; --write the data
					total_number_of_pixels <= total_number_of_pixels -1;
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_cs_idle_after_clearscreen;
					
				when st_cs_idle_after_clearscreen =>
					CS_ILI9325 <= '1'; --deselect the ili9325 chip
					if (total_number_of_pixels = X"0000") then --if all pixels are updated then we are finished
						ili9325_state <= st_delay; --to leave the screen in the color for 1 second after init is done I use a delay here
						delaytime_counter <= std_logic_vector(to_unsigned(Crystal_frequency, 32)); -- waits 1 second
						--delaytime_counter <= std_logic_vector(to_unsigned(250, 32)); --use this to shorten the simulation time
						nextstate <= st_wait_for_frame_completion; --now we wait for the current camera frame to complete 
					else
						delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
						ili9325_state <= st_delay; 
						nextstate <= st_prepare_clearscreen_cs; --go to the state to update the next pixel
					end if;					
					
				when st_delay => 
					delaytime_counter <= delaytime_counter -1;
					if (delaytime_counter = X"00") then
						ili9325_state <= nextstate; --when the delaytime has finished go to the state that was stored in "nextstate" before the delaytime state started
					end if;
					
-----------from here the state machine waits for a start of frame from the camera and if this occurs it takes care that all pixels of the frame are transferred to the TFT--------------------------
				--before st_idle is reached we wait in st_wait_for_frame_completion for the current frame to complete
				
				when st_idle => --previous frame has completed, in this state we wait until the NEW_FRAME_FOR_TFT signal goes low (which is connected to the VSYNC line of the OV7670)
					RD_ILI9325 <= '1'; --we do not use read so keep it inactive
					CS_ILI9325 <= '1'; --ili9325 not selected
					ILI9325_READY <= '1';
					ili9325_is_initialised <= '1';
					if (NEW_FRAME_FOR_TFT = '0') then --start of a new frame
						total_number_of_pixelbytes	<= std_logic_vector(to_unsigned(153600, 32)); --set the pixelbyte counter 320(H)x240(W)x2(Bytes per pixel)
						RS_ILI9325 <= '1'; --prepare RS pulse for write data
						CS_ILI9325 <= '0'; --select the display
						ili9325_state <= st_showframe_prepare_setXcursor_cs;
					end if;
				
				when st_showframe_prepare_setXcursor_cs =>	
					CS_ILI9325 <= '0'; --select the ili9325 chip
					DATA_TO_ILI9325 <= X"0020";
					ili9325_state <= st_delay;
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_prepare_setXcursor_rs_addresmode;
					
				when st_showframe_prepare_setXcursor_rs_addresmode =>	--set RS line for address/cmd mode
					RS_ILI9325 <= '0'; --prepare RS pulse for write address
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_prepare_setXcursor_wr;
					
				when st_showframe_prepare_setXcursor_wr =>	--prepare the wr line
					WR_ILI9325 <= '0'; --prepare the WR pulse
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_write_setXcursor_cmd;
					
				when st_showframe_write_setXcursor_cmd => --now write the command/address
					WR_ILI9325 <= '1'; --write the address
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_prepare_setXcursor_rs_datamode;	
					
				when st_showframe_prepare_setXcursor_rs_datamode => --prepare RS line for datamode
					RS_ILI9325 <= '1'; --prepare RS pulse for write data
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_prepare_write_setXcursordata;
					
				when st_showframe_prepare_write_setXcursordata => --prepare the data to send
					DATA_TO_ILI9325 <= X"0000";
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_prepare_setXcursor_data_wr;
					
				when st_showframe_prepare_setXcursor_data_wr =>
					WR_ILI9325 <= '0'; --prepare the WR pulse
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_write_setXcursor_data;
					
				when st_showframe_write_setXcursor_data =>	
					WR_ILI9325 <= '1'; --write the data
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_idle_after_setXcursor_data;
					
				when st_showframe_idle_after_setXcursor_data =>
					CS_ILI9325 <= '1'; --deselect the ili9325 chip
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_prepare_setYcursor_cs;
				

				when st_showframe_prepare_setYcursor_cs =>	
					CS_ILI9325 <= '0'; --select the ili9325 chip
					DATA_TO_ILI9325 <= X"0021";
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_prepare_setYcursor_rs_addresmode;
					
				when st_showframe_prepare_setYcursor_rs_addresmode =>	--set RS line for address/cmd mode
					RS_ILI9325 <= '0'; --prepare RS pulse for write address
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_prepare_setYcursor_wr;
					
				when st_showframe_prepare_setYcursor_wr =>	--prepare the wr line
					WR_ILI9325 <= '0'; --prepare the WR pulse
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(2, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_write_setYcursor_cmd;
					
				when st_showframe_write_setYcursor_cmd => --now write the command/address
					WR_ILI9325 <= '1'; --write the address
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_prepare_setYcursor_rs_datamode;	
					
				when st_showframe_prepare_setYcursor_rs_datamode => --prepare RS line for datamode
					RS_ILI9325 <= '1'; --prepare RS pulse for write data
					ili9325_state <= st_delay; delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_prepare_write_setYcursordata;
					
				when st_showframe_prepare_write_setYcursordata => --prepare the data to send
					DATA_TO_ILI9325 <= X"0000";
					ili9325_state <= st_delay;
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait					
					nextstate <= st_showframe_prepare_setYcursor_data_wr;
					
				when st_showframe_prepare_setYcursor_data_wr =>
					WR_ILI9325 <= '0'; --prepare the WR pulse
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_write_setYcursor_data;
					
				when st_showframe_write_setYcursor_data =>	
					WR_ILI9325 <= '1'; --write the data
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_idle_after_setYcursor_data;
					
				when st_showframe_idle_after_setYcursor_data =>
					CS_ILI9325 <= '1'; --deselect the ili9325 chip
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_prepare_select_gram_cs; --next state will prepare for a clear-screen of the TFT and put it in a certain color
					
					
				when st_showframe_prepare_select_gram_cs =>
					CS_ILI9325 <= '0'; --select the ili9325 chip
					DATA_TO_ILI9325 <= X"0022"; --select GRAM register;
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_prepare_select_gram_rs;
					
				when st_showframe_prepare_select_gram_rs =>
					RS_ILI9325 <= '0'; --prepare RS pulse for write address
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_prepare_select_gram_wr;
					
				when st_showframe_prepare_select_gram_wr =>
					WR_ILI9325 <= '0'; --prepare the WR pulse
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_showframe_select_gram;

				when st_showframe_select_gram =>
					WR_ILI9325 <= '1'; --write the address
					ili9325_state <= st_delay; 
					delaytime_counter <= std_logic_vector(to_unsigned(1, 32)); -- 1 clock period is 40nS, value after ( is the number of clockperiods to wait
					nextstate <= st_data2display_pixel_buffering_and_prepare_rs_and_cs;
					
---------------------at this point the image processing data must be transferred to the TFT--------------------------------------------------------------
---------------------the data is already in 16-bits, the only thing to do here is to put it on the TFT---------------------------------------------------
					
				when st_data2display_pixel_buffering_and_prepare_rs_and_cs =>
					RS_ILI9325 <= '1'; --prepare RS pulse for write data
					CS_ILI9325 <= '0'; --select the display
					if HREF_FROM_OV7670 = '1' then --there are still pixels coming out the camera!
							pixelbuffer1 <= CAMERA_DATA_IN;
							ili9325_state <= st_data2display_pixel_buffering_and_prepare_wr;
					end if;
					
				when st_data2display_pixel_buffering_and_prepare_wr =>
					WR_ILI9325 <= '0'; --prepare WR pulse for write data
					DATA_TO_ILI9325 <= pixelbuffer1; --prepare the data to send
					if HREF_FROM_OV7670 = '1' then 
						pixelbuffer2 <= CAMERA_DATA_IN; --and buffer the next pixel
						ili9325_state <= st_data2display_write_pixel;
					end if;	
					
				when st_data2display_write_pixel =>
					WR_ILI9325 <= '1'; --write the pixel to the TFT, the TFT address increments automatically
					total_number_of_pixelbytes <= total_number_of_pixelbytes -2; --increment pixelbyte counter (increment by 2 since every pixeltransfer is 2 bytes)
					pixelbuffer1 <= pixelbuffer2;--move the 2nd pixel to the first buffer so it will be put on the databus when returning to _buffering_and_prepare_wr
					if HREF_FROM_OV7670 = '1' then --there are still pixels coming out the camera!
						pixelbuffer2 <= CAMERA_DATA_IN;--store the 3rd pixel DVEDEBUG I have to see if this goes right since pixelbuffer2 is both read and assigned in 1 CLK cycle
					end if;
					if (total_number_of_pixelbytes = 0) then --finished! go idle and wait for a next camera frame
						ili9325_state <= st_wait_for_frame_completion;
					else
						ili9325_state <= st_data2display_pixel_buffering_and_prepare_wr;
					end if;
					
				when st_wait_for_frame_completion => --wait here for the OV7670 VSYNC line to go high/idle
					if (NEW_FRAME_FOR_TFT = '1') then --end of the frame
						ili9325_state <= st_idle;
					end if;
					
			end case;	
		end if;
	end process ili9325_statemachine;		


--The ILI9325 initialisation array, odd index numbers are register addresses, even index numbers contain the data for the register
	array_with_ili9325_initvalues(0) <= X"0000"; --address
	array_with_ili9325_initvalues(1) <= X"0001"; --data
	array_with_ili9325_initvalues(2) <= X"00e5"; --and so on....
	array_with_ili9325_initvalues(3) <= X"78f0";
	array_with_ili9325_initvalues(4) <= X"0001";
	array_with_ili9325_initvalues(5) <= X"0100";
	array_with_ili9325_initvalues(6) <= X"0002";
	array_with_ili9325_initvalues(7) <= X"0700";
	array_with_ili9325_initvalues(8) <= X"0003";
	array_with_ili9325_initvalues(9) <= X"1038";--YES! X"1038" gives an image of the ov7670 on my ili9325 TFT! This one is important and can arrange the display view in just about any way so experiment if the display is not showing correctly
	array_with_ili9325_initvalues(10) <= X"0004";
	array_with_ili9325_initvalues(11) <= X"0000";
	array_with_ili9325_initvalues(12) <= X"0008";
	array_with_ili9325_initvalues(13) <= X"0202";
	array_with_ili9325_initvalues(14) <= X"0009";
	array_with_ili9325_initvalues(15) <= X"0000";
	array_with_ili9325_initvalues(16) <= X"000a";
	array_with_ili9325_initvalues(17) <= X"0000";
	array_with_ili9325_initvalues(18) <= X"000c";
	array_with_ili9325_initvalues(19) <= X"0001";
	array_with_ili9325_initvalues(20) <= X"000d";
	array_with_ili9325_initvalues(21) <= X"0000";
	array_with_ili9325_initvalues(22) <= X"000f";
	array_with_ili9325_initvalues(23) <= X"0000";
	array_with_ili9325_initvalues(24) <= X"0010";
	array_with_ili9325_initvalues(25) <= X"0000";
	array_with_ili9325_initvalues(26) <= X"0011";
	array_with_ili9325_initvalues(27) <= X"0007";
	array_with_ili9325_initvalues(28) <= X"0012";
	array_with_ili9325_initvalues(29) <= X"0000";
	array_with_ili9325_initvalues(30) <= X"0013";
	array_with_ili9325_initvalues(31) <= X"0000";
	array_with_ili9325_initvalues(32) <= X"0007";
	array_with_ili9325_initvalues(33) <= X"0001";
	array_with_ili9325_initvalues(34) <= X"0010";
	array_with_ili9325_initvalues(35) <= X"1690";
	array_with_ili9325_initvalues(36) <= X"0011";
	array_with_ili9325_initvalues(37) <= X"0227";
	array_with_ili9325_initvalues(38) <= X"0012";
	array_with_ili9325_initvalues(39) <= X"009d";
	array_with_ili9325_initvalues(40) <= X"0013";
	array_with_ili9325_initvalues(41) <= X"1900";
	array_with_ili9325_initvalues(42) <= X"0029";
	array_with_ili9325_initvalues(43) <= X"0025";
	array_with_ili9325_initvalues(44) <= X"002b";
	array_with_ili9325_initvalues(45) <= X"000d";
	array_with_ili9325_initvalues(46) <= X"0020";
	array_with_ili9325_initvalues(47) <= X"0000";
	array_with_ili9325_initvalues(48) <= X"0021";
	array_with_ili9325_initvalues(49) <= X"0000";
	array_with_ili9325_initvalues(50) <= X"0030";
	array_with_ili9325_initvalues(51) <= X"0007";
	array_with_ili9325_initvalues(52) <= X"0031";
	array_with_ili9325_initvalues(53) <= X"0303";
	array_with_ili9325_initvalues(54) <= X"0032";
	array_with_ili9325_initvalues(55) <= X"0003";
	array_with_ili9325_initvalues(56) <= X"0035";
	array_with_ili9325_initvalues(57) <= X"0206";
	array_with_ili9325_initvalues(58) <= X"0036";
	array_with_ili9325_initvalues(59) <= X"0008";
	array_with_ili9325_initvalues(60) <= X"0037";
	array_with_ili9325_initvalues(61) <= X"0406";
	array_with_ili9325_initvalues(62) <= X"0038";
	array_with_ili9325_initvalues(63) <= X"0304";
	array_with_ili9325_initvalues(64) <= X"0039";
	array_with_ili9325_initvalues(65) <= X"0007";
	array_with_ili9325_initvalues(66) <= X"003c";
	array_with_ili9325_initvalues(67) <= X"0602";
	array_with_ili9325_initvalues(68) <= X"003d";
	array_with_ili9325_initvalues(69) <= X"0008";
	array_with_ili9325_initvalues(70) <= X"0050";
	array_with_ili9325_initvalues(71) <= X"0000";
	array_with_ili9325_initvalues(72) <= X"0051";
	array_with_ili9325_initvalues(73) <= X"00ef";--X"00ef";
	array_with_ili9325_initvalues(74) <= X"0052";
	array_with_ili9325_initvalues(75) <= X"0000";
	array_with_ili9325_initvalues(76) <= X"0053";
	array_with_ili9325_initvalues(77) <= X"013f";--X"013f";
	array_with_ili9325_initvalues(78) <= X"0060";
	array_with_ili9325_initvalues(79) <= X"a700";--was 2700
	array_with_ili9325_initvalues(80) <= X"0061";
	array_with_ili9325_initvalues(81) <= X"0001";
	array_with_ili9325_initvalues(82) <= X"006a";
	array_with_ili9325_initvalues(83) <= X"0000";
	array_with_ili9325_initvalues(84) <= X"0080";
	array_with_ili9325_initvalues(85) <= X"0000";
	array_with_ili9325_initvalues(86) <= X"0081";
	array_with_ili9325_initvalues(87) <= X"0000";
	array_with_ili9325_initvalues(88) <= X"0082";
	array_with_ili9325_initvalues(89) <= X"0000";
	array_with_ili9325_initvalues(90) <= X"0083";
	array_with_ili9325_initvalues(91) <= X"0000";
	array_with_ili9325_initvalues(92) <= X"0084";
	array_with_ili9325_initvalues(93) <= X"0000";
	array_with_ili9325_initvalues(94) <= X"0085";
	array_with_ili9325_initvalues(95) <= X"0000";
	array_with_ili9325_initvalues(96) <= X"0090";
	array_with_ili9325_initvalues(97) <= X"0010";
	array_with_ili9325_initvalues(98) <= X"0092";
	array_with_ili9325_initvalues(99) <= X"0600";
	array_with_ili9325_initvalues(100) <= X"0007";
	array_with_ili9325_initvalues(101) <= X"0133"; --total of 102 values including index 0
end Behavioral;




