-----------------------------------------------------------------------------------------------------------------------------------
-- Engineer			: 		Dennis van Elteren, Maarssen, The Netherlands
-- 
-- Create Date		:    	20:23:22 05/13/2015 
-- Design Name		: 		My FPGA vision system
-- Module Name		:    	the image tresholding module
-- Target Devices	: 		xc3s200,xc6slx9
-- Tool versions	: 		ISE14.7
-- Description		: 		This vhdl code gathers pixels from an Omnivision OV7670 camera, combines it to 16-bit words
--								and does a simple treshold image processing action on it: if R,G,B values of the 16-bit word match a certain
--								criterium they are labeled white, otherwise they are labeled black. The camera gives data in RGB565 mode meaning
--								that the R-value = 5 bit, G value = 6 bit, B value = 5 bit and each pixel comes in 2 bytes
--								the processed data is output as 16-bit words and since the processing takes a number of CLK cycles the VSYNC and HREF
--							 	lines are also "extended" to the amount of cycles it took before the first processed pixel comes out
--
-- Current Rev		: 		1.1	-	18-5-2015
-- History					: 		1.0	- 	File Created
--								:		1.1	-	First working version, can treshold an image and so find "bright" spots in an image
--	TODO				: 		
---------------------------------------------------------------------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL; --for casting std_logic_vector to unsigned

entity image_tresholding is
	 Generic (Crystal_frequency : integer   := 25000000; --the constants
				 red_treshold      : natural     := 27;
				 green_treshold    : natural     := 59;
				 blue_treshold     : natural     := 27;
				 vsync_extension_clk_cycles    : integer   := 3; --this value is calculated by counting the number of states of the processing statemachine it takes before actual data arrives to the output
				 href_extension_clk_cycles     : integer   := 3);--same for this value
    Port ( VSYNC_IN 			   : in  STD_LOGIC;
           HREF_IN 			   : in  STD_LOGIC;
           CAMERA_DATABUS_IN  : in  STD_LOGIC_VECTOR (7 downto 0);
           CLOCK_IN 			   : in  STD_LOGIC;
			  TRESHOLD_OR_NORMAL	: in  STD_LOGIC;
           HREF_OUT 			   : out  STD_LOGIC := '0';
           VSYNC_OUT 		   : out  STD_LOGIC := '1';
           CAMERA_DATABUS_OUT : out  STD_LOGIC_VECTOR (15 downto 0) := X"0000");
end image_tresholding;

architecture Behavioral of image_tresholding is

--these counters are used to "shift" the VSYNC low time a number of clk cycles and do the same for the HREF high time
signal vsync_shifter							: std_logic_vector(3 downto 0) := std_logic_vector(to_unsigned(vsync_extension_clk_cycles, 4)); 
signal href_shifter							: std_logic_vector(3 downto 0) := std_logic_vector(to_unsigned(href_extension_clk_cycles, 4)); 

signal vsync_out_state : std_logic := '1'; --signal is used to compare VSYNC_IN with VSYNC_OUT, output cannot be read so its done via a signal
signal href_out_state  : std_logic := '0'; --same trick for href
signal total_number_of_pixelbytes	: std_logic_vector(31 downto 0) := std_logic_vector(to_unsigned(153600, 32)); --total nr of pixelbytes coming from the OV7670

--statemachine which does the pixel tresholding
type treshold_states is (st_wait_for_vsync_low,st_capture_pixel1_lsb,st_capture_pixel1_msb,st_capture_pixel2_lsb_and_tresholding,st_capture_pixel2_msb,st_wait_for_frame_completion);

-- Register to store the current state(s) or nextstate to go to
signal treshold_state : treshold_states := st_wait_for_vsync_low; --intial state 

--signals to store temporary data during the running of operating state machine
signal pixel1	: std_logic_vector(15 downto 0) := X"0000";
signal pixel2	: std_logic_vector(15 downto 0) := X"0000";

signal tresholding : std_logic := '1';

--debugging signals
signal extractedRED 		: std_logic_vector(4 downto 0);
signal extractedGREEN 	: std_logic_vector(5 downto 0);
signal extractedBLUE 	: std_logic_vector(4 downto 0);
signal debugRED			: std_logic_vector(4 downto 0) := "11100";

begin
	
	treshold_statemachine :	process(CLOCK_IN)
	begin
		if rising_edge(CLOCK_IN) then
			case treshold_state is
			
				when st_wait_for_vsync_low => --wait for a new frame from the camera
					if (VSYNC_IN = '0') then 
						total_number_of_pixelbytes	<= std_logic_vector(to_unsigned(153600, 32)); --set the pixelbyte counter 320(H)x240(W)x2(Bytes per pixel)
						treshold_state <= st_capture_pixel1_lsb;
					end if;
					
				when st_capture_pixel1_lsb =>
					if HREF_IN = '1' then
						pixel1(7 downto 0) <= CAMERA_DATABUS_IN; --get the first (ls)byte of the first pixel clocked in after HREF went active
						total_number_of_pixelbytes <= total_number_of_pixelbytes -1; --increment pixelbyte counter
						treshold_state <= st_capture_pixel1_msb;
					end if;
					
				when st_capture_pixel1_msb =>
					if HREF_IN = '1' then
						pixel1(15 downto 8) <= CAMERA_DATABUS_IN; --get the second (ms)byte of the first pixel clocked in after HREF went active
						total_number_of_pixelbytes <= total_number_of_pixelbytes -1; --increment pixelbyte counter
						treshold_state <= st_capture_pixel2_lsb_and_tresholding;
					end if;
				
				when st_capture_pixel2_lsb_and_tresholding =>
					if HREF_IN = '1' then
						pixel2(7 downto 0) <= CAMERA_DATABUS_IN; --get the first (ls)byte of the second pixel clocked in after HREF went active
						total_number_of_pixelbytes <= total_number_of_pixelbytes -1; --increment pixelbyte counter
						treshold_state <= st_capture_pixel2_msb;
					end if;
					if tresholding = '0' then  
						CAMERA_DATABUS_OUT <= pixel1; --show the normal image from the camera
					else --do pixel tresholding and show this on the TFT:
						if( (to_integer(unsigned(pixel1(15 downto 11))) > red_treshold) AND (to_integer(unsigned(pixel1(10 downto 5))) > green_treshold) AND (to_integer(unsigned(pixel1(4 downto 0))) > blue_treshold) ) then 	
							CAMERA_DATABUS_OUT <= X"ffff";
						else
							CAMERA_DATABUS_OUT <= X"0000";
						end if;
					end if;
				
				when st_capture_pixel2_msb =>
					if HREF_IN = '1' then
						pixel1(15 downto 8) <= CAMERA_DATABUS_IN; --get the second (ms)byte of the second pixel clocked in after HREF went active,store in pixel1 MSB since this is free to use again in this state
						pixel1(7 downto 0) <= pixel2(7 downto 0);--move the lsbyte just captured in pixel2 signal to pixel1 signal
						total_number_of_pixelbytes <= total_number_of_pixelbytes -1; --increment pixelbyte counter
					end if;
					 
					if (total_number_of_pixelbytes = 1) then --finished! go idle and wait for a next camera frame
						treshold_state <= st_wait_for_frame_completion;
					else
						treshold_state <= st_capture_pixel2_lsb_and_tresholding; --do the tresholding again on the next pixel, keep storing new pixels
					end if;
					
				when st_wait_for_frame_completion => --wait here for the OV7670 VSYNC line to go high/idle
					if (VSYNC_IN = '1') then --end of the frame
						treshold_state <= st_wait_for_vsync_low; --start over
					end if;
				
			end case;			
		end if;
	end process treshold_statemachine;
	
	--process to monitor the digilent board SW0, depending on its value the view on the TFT is nomal or tresholded
	treshold_or_normal_view :	process(CLOCK_IN)
		begin
			if rising_edge(CLOCK_IN) then
				if TRESHOLD_OR_NORMAL = '0' then 
					tresholding	 <= '0';
				else
					tresholding	 <= '1';
				end if;
			end if;
	end process treshold_or_normal_view;
	
	--this process delays the start of VSYNC and extends the VSYNC pulse after the original VSYNC has gone back to idle (HIGH)
	vsync_shift :	process(CLOCK_IN)
		begin
			if rising_edge(CLOCK_IN) then
				if (VSYNC_IN = NOT vsync_out_state) then --this is the case when the input has changed!
					vsync_shifter	 <= vsync_shifter	 -1;
				end if;
					
				if (vsync_shifter	 = X"0000") then
					VSYNC_OUT <= VSYNC_IN;
					vsync_out_state <= VSYNC_IN; --to "shadow" the status of the output, the current process cannot read from an output unless defined as inout which is overkill
					vsync_shifter <= std_logic_vector(to_unsigned(vsync_extension_clk_cycles, 4)); --reset the extention period counter 
				end if;
			end if;
	end process vsync_shift;
	
	--this process delays the start of HREF and extends the HREF pulse after the original HREF has gone back to idle (LOW)
	href_shift :	process(CLOCK_IN)
		begin
			if rising_edge(CLOCK_IN) then
				if (HREF_IN = NOT href_out_state) then --this is the case when the input has changed!
					href_shifter	 <= href_shifter	-1;
				end if;
					
				if (href_shifter	 = X"0000") then
					HREF_OUT <= HREF_IN;
					href_out_state <= HREF_IN; --to "shadow" the status of the output, the current process cannot read from an output unless defined as inout which is overkill
					href_shifter <= std_logic_vector(to_unsigned(href_extension_clk_cycles, 4)); --reset the extention period counter 
				end if;
			end if;
	end process href_shift;
				
end Behavioral;

