------------------------------------------------------------------------------------------------------------------------------------------------------
-- Engineer			: 		Dennis van Elteren, Maarssen, The Netherlands
-- 
-- Create Date		:    	February 24th 2015 
-- Design Name		: 		My FPGA vision system
-- Module Name		:    	ov7670_sccb_initialisation
-- Target Devices	: 		xc3s200,xc6slx9
-- Tool versions	: 		ISE14.7
-- Description		: 		This vhdl code initialises an Omnivision OV7670 camera via it's SCCB interface
--								After initialisation the camera is in QVGA (320x240) and RGB565 mode.
--								INIT_DONE is then HIGH which signals to the toplevel to connect the camera's VSYNC 
--								to the ili9325 code block. Now the TFT with ili9325 will show the images coming from the OV7670
--								The SCCB interface is "fire and forget" meaning that the status after sending a byte is read back (9th SCL dummy required for sccb)
--								The read-back (dummy clk cycle) is mandatory but on the result of it is not acted on.
--								This implies that if an error occurs during the SCCB initialisation sequence of the ov7670 it will not be noticed and the camera
--								might be wrong or half initialised. I accepted this risk. given the effort it takes to implement the reading back.
-- Remarks			:
--
-- Current Rev		: 		1.1
-- History			: 		1.0	- 	File Created
--								1.1	-	First version which is working!
--	TODO				: 		
---------------------------------------------------------------------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL; --for casting std_logic_vector to unsigned

entity ov7670_sccb_initialisation is
    Generic (Crystal_frequency : integer   := 25000000; --to shorten all delays in simulation: make this value smaller
				 totalarraysize    : integer   := 48); -- size of the initialisation array (starting from index 0...end)
	 Port ( CLOCK_IN 		: in  STD_LOGIC;
           SDA_IN 		: in  STD_LOGIC;
			  ENABLE			: in	STD_LOGIC;
			  NOT_RESET_IN	: in 	STD_LOGIC;
           SDA_OUT 		: out  STD_LOGIC := '1';
           SCL_OUT 		: out  STD_LOGIC := '1';
           INIT_DONE 	: out  STD_LOGIC := '0');
end ov7670_sccb_initialisation;

architecture Behavioral of ov7670_sccb_initialisation is

signal delaytime_counter 				: std_logic_vector(31 downto 0) := std_logic_vector(to_unsigned((Crystal_frequency/2000), 32)); --Delay: 10000=100uS,2000=500uS,200=5mS,20=50mS,10=100ms
signal bitcounter							: std_logic_vector(3 downto 0) := std_logic_vector(to_unsigned(8, 4)); --takes care that each sccb write puts out 8 bits

type sccb_state is (st_reset,st_wait_for_enable,st_start_sccb,st_sccbstart_sda_H,st_sccbstart_sda_L,st_sccbstart_scl_H,st_sccbstart_scl_L,st_prepare_send_write_address,st_send_write_address,
						  st_send_write_address_scl_H,st_send_write_address_scl_L,st_write_address_dummy_scl_H,st_write_address_dummy_scl_L,st_prepare_send_regID,st_send_regID,st_send_regID_scl_H,st_send_regID_scl_L,
						  st_send_regID_dummy_scl_H,st_send_regID_dummy_scl_L,st_prepare_send_regDATA,st_send_regDATA,st_send_regDATA_scl_H,st_send_regDATA_scl_L,st_send_regDATA_dummy_scl_H,st_send_regDATA_dummy_scl_L,
						  st_stop_sccb,st_sccbstop_sda_L,st_sccbstop_scl_H,st_sccbstop_sda_H,st_repeat_loop_or_done,st_delay5ms,st_delay50ms,st_delay500us,st_delay100us,st_done);

-- Register to store the current state(s) or nextstate to go to
signal ov7670_state : sccb_state := st_reset; --intial state 
--signal ov7670_state : sccb_state := st_start_sccb;--use this for simulation to skip the 100mS reset
signal nextstate : sccb_state := st_reset; --this signal will be used to jump to the next desired state from a previous state, this way states can be re-used

--array and signals to initialise the ov7670 camera
type initarray is array (0 to totalarraysize) of std_logic_vector(7 downto 0); --the array containing the initvalues is below all the VHDL code
signal array_with_ov7670_initvalues : initarray;   --instantiate 1 array of the previously defined type
signal index : integer range 0 to totalarraysize;

signal ov7670_wr_address : std_logic_vector(7 downto 0) := X"42"; --sccb (I2C) write address of the ov7670
signal ov7670_rd_address : std_logic_vector(7 downto 0) := X"43"; --sccb (I2C) read address of the ov7670
signal sccb_databyte		 : std_logic_vector(7 downto 0) := X"00"; --data to be sent out via sccb will be stored here

begin
	
	ov7670_statemachine :	process(CLOCK_IN)
	begin
		if rising_edge(CLOCK_IN) then
			case ov7670_state is
				when st_reset => --wait for a number of clk cycles (power on reset)
					if (NOT_RESET_IN = '1') then 
						ov7670_state <= st_wait_for_enable;
					end if;
				
				when st_wait_for_enable =>
					if (ENABLE = '1') then
						ov7670_state <= st_start_sccb;
					end if;
----------------------------step 1: START SCCB-----------------------------------------------------------------				
				when st_start_sccb => --this is just a dummy state to make the state machine more readable
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32)); --Delay: 10000=100uS,2000=500uS,200=5mS,20=50mS 
					ov7670_state <= st_delay500us; --wait a while and after that... 
					nextstate <= st_sccbstart_sda_H; --....go to the next state which is sda_H
					
				when st_sccbstart_sda_H => 
					SDA_OUT <= '1';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32)); --Delay: 10000=100uS,2000=500uS,200=5mS,20=50mS 
					ov7670_state <= st_delay500us; 
					nextstate <= st_sccbstart_scl_H;
				
				when st_sccbstart_scl_H =>
					SCL_OUT <= '1';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32)); --Delay: 10000=100uS,2000=500uS,200=5mS,20=50mS 
					ov7670_state <= st_delay500us; 
					nextstate <= st_sccbstart_sda_L;	
					
				when st_sccbstart_sda_L =>
					SDA_OUT <= '0';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32)); --Delay: 10000=100uS,2000=500uS,200=5mS,20=50mS 
					ov7670_state <= st_delay500us; 
					nextstate <= st_sccbstart_scl_L;
					
				when st_sccbstart_scl_L =>
					SCL_OUT <= '0';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32)); --Delay: 10000=100uS,2000=500uS,200=5mS,20=50mS 
					ov7670_state <= st_delay500us; 
					nextstate <= st_prepare_send_write_address;
					
-------------------------end step 1: START SCCB-------------------------------------------------------------------------------
-------------------------step 2: send write address-------------------------------------------------------------------	
				
				when st_prepare_send_write_address =>
					--SCL_OUT <= '0'; --to be sure SCL is LOW
					sccb_databyte <= ov7670_wr_address;
					bitcounter <= std_logic_vector(to_unsigned(8, 4)); --reset the bitcounter
					ov7670_state <= st_send_write_address;
					
				when st_send_write_address =>
					bitcounter <= bitcounter -1;
					SDA_OUT <= sccb_databyte(7); --sending out MS-bit
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us;
					nextstate <= st_send_write_address_scl_H;
					
				when st_send_write_address_scl_H =>
					SCL_OUT <= '1';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us; 
					nextstate <= st_send_write_address_scl_L;
					
				when st_send_write_address_scl_L =>
					SCL_OUT <= '0';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us; 
					if (bitcounter = X"00") then
						nextstate <= st_write_address_dummy_scl_H; --this is required by the sccb standard, logic 1 on sda line indicates a transmit faillure but I do not check on it
					else
						sccb_databyte <= sccb_databyte(6 downto 0) & '0';--shift left so the next bit will be sent
						nextstate <= st_send_write_address;
					end if;
					
				when st_write_address_dummy_scl_H => --9th scl rising edge, dummy action required for the sccb protocol
					SCL_OUT <= '1';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us;
					nextstate <= st_write_address_dummy_scl_L;
					
				when st_write_address_dummy_scl_L => --9th scl falling edge, dummy action required for the sccb protocol
					SCL_OUT <= '0';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us;
					nextstate <= st_prepare_send_regID;
					
---------------------end step 2: send write address-------------------------------------------------------------------
---------------------step 3: send the regID---------------------------------------------------------------------------

				when st_prepare_send_regID =>
					sccb_databyte <= array_with_ov7670_initvalues(index);
					bitcounter <= std_logic_vector(to_unsigned(8, 4)); --reset the bitcounter
					ov7670_state <= st_send_regID;
					
				when st_send_regID =>
					bitcounter <= bitcounter -1;
					SDA_OUT <= sccb_databyte(7); --sending out MS-bit
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us;
					nextstate <= st_send_regID_scl_H;
					
				when st_send_regID_scl_H =>
					SCL_OUT <= '1';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us; 
					nextstate <= st_send_regID_scl_L;
					
				when st_send_regID_scl_L =>
					SCL_OUT <= '0';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us; 
					if (bitcounter = X"00") then
						nextstate <= st_send_regID_dummy_scl_H; --like with send_write_address: a 9th dummy scl rising edge is required
						index <= index + 1; --next array variable contains the data
					else
						sccb_databyte <= sccb_databyte(6 downto 0) & '0';--shift left so the next bit will be sent
						nextstate <= st_send_regID;
					end if;
				
				when st_send_regID_dummy_scl_H => --9th scl rising edge, dummy action required for the sccb protocol
					SCL_OUT <= '1';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us;
					nextstate <= st_send_regID_dummy_scl_L;
					
				when st_send_regID_dummy_scl_L => --9th scl rising edge, dummy action required for the sccb protocol
					SCL_OUT <= '0';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us;
					nextstate <= st_prepare_send_regDATA;
					
---------------------end step 3: send regID----------------------------------------------------------------------------------------
---------------------step 4: send the regDATA--------------------------------------------------------------------------------------

				when st_prepare_send_regDATA =>
					sccb_databyte <= array_with_ov7670_initvalues(index);
					bitcounter <= std_logic_vector(to_unsigned(8, 4)); --reset the bitcounter
					ov7670_state <= st_send_regDATA;
					
				when st_send_regDATA =>
					bitcounter <= bitcounter -1;
					SDA_OUT <= sccb_databyte(7); --sending out MS-bit
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us;
					nextstate <= st_send_regDATA_scl_H;
					
				when st_send_regDATA_scl_H =>
					SCL_OUT <= '1';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us; 
					nextstate <= st_send_regDATA_scl_L;
					
				when st_send_regDATA_scl_L =>
					SCL_OUT <= '0';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us; 
					if (bitcounter = X"00") then
						index <= index + 1; --next array variable contains the data
						nextstate <= st_send_regDATA_dummy_scl_H; --same here as send_write_address and send_regID: a 9th dummy scl rising edge is required
					else
						sccb_databyte <= sccb_databyte(6 downto 0) & '0';--shift left so the next bit will be sent
						nextstate <= st_send_regDATA;
					end if;
				
				when st_send_regDATA_dummy_scl_H => --9th scl rising edge, dummy action required for the sccb protocol
					SCL_OUT <= '1';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us;
					nextstate <= st_send_regDATA_dummy_scl_L;
					
				when st_send_regDATA_dummy_scl_L => --9th scl rising edge, dummy action required for the sccb protocol
					SCL_OUT <= '0';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us;
					nextstate <= st_stop_sccb;
					
---------------------end step 4: send regData----------------------------------------------------------------------------------------
----------------------------step 5: STOP SCCB-----------------------------------------------------------------				
				when st_stop_sccb => 
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us; 
					nextstate <= st_sccbstop_sda_L; 
				
				when st_sccbstop_sda_L =>
					SDA_OUT <= '0';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us; 
					nextstate <= st_sccbstop_scl_H;
					
				when st_sccbstop_scl_H =>
					SCL_OUT <= '1';
					delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
					ov7670_state <= st_delay500us; 
					nextstate <= st_sccbstop_sda_H;
					
				when st_sccbstop_sda_H =>
					SDA_OUT <= '1';
					if(index = 2) then --this means the SCCB reset command was sent so we need to wait longer, until the reset is finished
						delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/20), 32));
						ov7670_state <= st_delay50ms;
					else
						delaytime_counter <= std_logic_vector(to_unsigned((Crystal_frequency/2000), 32));
						ov7670_state <= st_delay500us; 
					end if;
					nextstate <= st_repeat_loop_or_done;
					
-------------------------end step 5: STOP SCCB-------------------------------------------------------------------------------
-------------------------end step 6: repeat the loop or are we done?---------------------------------------------------------
				
				when st_repeat_loop_or_done =>
					if(index = totalarraysize) then --is the complete array sent out? then we are done
						ov7670_state <= st_done;
					else
						ov7670_state <= st_start_sccb; --restart sccb and send the next regID and regDAT values
					end if;
-------------------------end step 6: repeat the loop or are we done----------------------------------------------------------				
				when st_delay100us =>
					delaytime_counter <= delaytime_counter -1;
					if (delaytime_counter = X"00") then
						ov7670_state <= nextstate;
					end if;	
				
				when st_delay500us =>
					delaytime_counter <= delaytime_counter -1;
					if (delaytime_counter = X"00") then
						ov7670_state <= nextstate;
					end if;	
					
				when st_delay5ms =>
					delaytime_counter <= delaytime_counter -1;
					if (delaytime_counter = X"00") then
						ov7670_state <= nextstate;
					end if;	
				
				when st_delay50ms =>
					delaytime_counter <= delaytime_counter -1;
					if (delaytime_counter = X"00") then
						ov7670_state <= nextstate;
					end if;
					
				when st_done =>
					INIT_DONE <= '1';
					
			end case;			
		end if;
	end process ov7670_statemachine;	

--ov7670 set to qvga rgb565, pclk = xclk
--below is the initialisation array, odd index numbers are register ID's, even index numbers contain the register data 
--first 2 values 0 and 1 take care that the SCCB interface is reset (0x12,0x80), this is mandatory
-------------------minimal settings version-------------------------------------------------------------------------------------------------
	array_with_ov7670_initvalues(0) <= X"12"; --first do an sccb reset
	array_with_ov7670_initvalues(1) <= X"80"; --reset sccb
	array_with_ov7670_initvalues(2) <= X"12"; --COM7
	array_with_ov7670_initvalues(3) <= X"14"; --0x14 QVGA, RGB mode
	array_with_ov7670_initvalues(4) <= X"8c"; --RGB444 register
	array_with_ov7670_initvalues(5) <= X"00"; --RGB444 off
	array_with_ov7670_initvalues(6) <= X"04"; --COM1
	array_with_ov7670_initvalues(7) <= X"00"; --CCIR656 disabled
	array_with_ov7670_initvalues(8) <= X"40"; --COM15
	array_with_ov7670_initvalues(9) <= X"10"; --0x10 RGB565 selected (RGB444 needs to be off, done with reg 0x8c)
	array_with_ov7670_initvalues(10) <= X"14"; --COM9
	array_with_ov7670_initvalues(11) <= X"18"; --automatic gain ceiling = 4
	array_with_ov7670_initvalues(12) <= X"11"; --CLKRC
	array_with_ov7670_initvalues(13) <= X"00"; --external clk is internal clk (25MHz)
	array_with_ov7670_initvalues(14) <= X"6b"; --DBLV
	array_with_ov7670_initvalues(15) <= X"40"; --input clk x 4 according to datasheet but these settings give me an internal (PCLK) clock of 25MHz)
	array_with_ov7670_initvalues(16) <= X"0c"; --COM3
	array_with_ov7670_initvalues(17) <= X"00"; --no scaling, DCW or byte swap
	array_with_ov7670_initvalues(18) <= X"32"; --HREF
	array_with_ov7670_initvalues(19) <= X"b6"; --HREF edge offset, end and start (I did not touch!)
	array_with_ov7670_initvalues(20) <= X"17"; --HSTART
	array_with_ov7670_initvalues(21) <= X"13"; --untouched, from the internet examples
	array_with_ov7670_initvalues(22) <= X"18"; --HSTOP
	array_with_ov7670_initvalues(23) <= X"01"; --untouched, from the internet examples
	array_with_ov7670_initvalues(24) <= X"19"; --VSTART
	array_with_ov7670_initvalues(25) <= X"02"; --untouched, from the internet examples
	array_with_ov7670_initvalues(26) <= X"1a"; --VSTOP
	array_with_ov7670_initvalues(27) <= X"7a"; ----untouched, from the internet examples
	array_with_ov7670_initvalues(28) <= X"03"; --VREF
	array_with_ov7670_initvalues(29) <= X"0a"; --untouched, from the internet examples
	array_with_ov7670_initvalues(30) <= X"3e"; --COM14
	array_with_ov7670_initvalues(31) <= X"00"; --no scaling no pixelclock divisions
	array_with_ov7670_initvalues(32) <= X"70"; --SCALING_XSC and testpattern on off
	array_with_ov7670_initvalues(33) <= X"3a"; --make X"3a",testpattern on is supposed to be X"BA" but with  3A and reg 0x71 set to B5 I also got a (perfect) testpattern;
	array_with_ov7670_initvalues(34) <= X"71"; --SCALING_YSC and testpattern on off
	array_with_ov7670_initvalues(35) <= X"35"; --was X"35" for testpattern on make it X"B5";
	array_with_ov7670_initvalues(36) <= X"b0"; --without this setting the pictures have no color and red = green
	array_with_ov7670_initvalues(37) <= X"84"; --reg b0 and value 84 come from the internet, the datasheet only sais RESERVED!!
	array_with_ov7670_initvalues(38) <= X"1e"; --flipping/mirroring the image
	array_with_ov7670_initvalues(39) <= X"11"; --my TFT needs vertical flipping (bit 1 and 0 are reserved, I kept them 0 and 1)
	-------------------END minimal settings version---------------------------------------------------------------------------------------------
	--until this point: totalarraysize    : integer   := 40
	
	--below here is used to debug problems or enhance the camera view
	--brightness experiment (reg 0x55), default value = 0x00
	array_with_ov7670_initvalues(40) <= X"55";
	array_with_ov7670_initvalues(41) <= X"08"; --0x08 gives good results
	
	--switch off all automatic enhancements and stuff to have a steady image with no automatic white balancing etc etc
	array_with_ov7670_initvalues(42) <= X"13"; --AGC,AWB,AEC all off
	array_with_ov7670_initvalues(43) <= X"00";
	array_with_ov7670_initvalues(44) <= X"3b"; --disable night mode, banding filters etc
	array_with_ov7670_initvalues(45) <= X"00";
	array_with_ov7670_initvalues(46) <= X"3d";
	array_with_ov7670_initvalues(47) <= X"c0";
end Behavioral;
	


